from typing import Iterable, Iterator, Optional, Tuple, TypeVar, Union

import operator

from itertools import chain
from itertools import islice
from itertools import product
from itertools import repeat
from itertools import starmap

import numpy as np

X = TypeVar("X")

Coordinate = Tuple[int, ...]
RawSlice = Union[
  Tuple[Optional[int]],
  Tuple[Optional[int], Optional[int]],
  Tuple[Optional[int], Optional[int], Optional[int]]
]
Shape = Tuple[int, ...]

def slices(xs: Iterable[RawSlice]) -> Tuple[slice, ...]:
  ys = starmap(slice, xs)
  return tuple(ys)

def create_batches(xs: Iterable[Tuple[np.ndarray, ...]], length: int = 32) -> Iterable[Tuple[np.ndarray, ...]]:
  xs = iter(xs)
  batch: Tuple[np.ndarray, ...] = next_batch(xs, length)

  while batch:
    yield batch
    batch: Tuple[np.ndarray, ...] = next_batch(xs, length)

def next_batch(xs: Iterator[Tuple[np.ndarray, ...]], length: int = 32) -> Tuple[np.ndarray, ...]:
  batch = islice(xs, length)
  stacked_batch = map(np.stack, zip(*batch))

  return tuple(stacked_batch)

def cycle_tensor(tensor: Union[np.ndarray, float], shape: Shape) -> np.ndarray:
  tensor = tensor if isinstance(tensor, np.ndarray) else np.array(tensor)
  expanded: np.ndarray = expand_tail_dimensions(tensor, len(shape))

  for axis, axis_length in enumerate(shape):
    cycled_axis: Tuple[np.ndarray, ...] = expand_tail((expanded,), axis_length, filler=expanded)
    expanded = np.concatenate(cycled_axis, axis=axis)

  return expanded

def expand_tail_dimensions(tensor: np.ndarray, length: int) -> np.ndarray:
  expanded_shape: Shape = expand_tail(tensor.shape, length, filler=1)
  return tensor.reshape(expanded_shape)

def expand_tail(xs: Tuple[X, ...], length: int, filler: X) -> Tuple[X, ...]:
  count = max(length - len(xs), 0)

  fillers = repeat(filler)
  sliced_fillers = islice(fillers, count)

  expanded = chain(xs, sliced_fillers)
  return tuple(expanded)

def gridrange(start: Coordinate, end: Optional[Coordinate] = None, step: Optional[Coordinate] = None) -> Iterable[Coordinate]:
  if end is None:
    end = start
    start = expand_tail((0,), len(end), filler=0)

  step = step or expand_tail((1,), len(start), filler=1)

  ranges = map(range, start, end, step)
  return product(*ranges)

def iteraxis(tensor: np.ndarray, axis: int = 0) -> Iterable[np.ndarray]:
  transposed = np.moveaxis(tensor, axis, -1)
  return map(lambda x: transposed[x], np.ndindex(transposed.shape[:-1]))

def move_tensor(source: np.ndarray, destination: np.ndarray, coordinate: Optional[Coordinate] = None) -> np.ndarray:
  tensor = destination.copy()
  move_tensor_inplace(source, tensor, coordinate)

  return tensor

def move_tensor_inplace(source: np.ndarray, destination: np.ndarray, coordinate: Optional[Coordinate] = None):
  coordinate = coordinate or (0,)

  start: Coordinate = expand_tail(coordinate, source.ndim, filler=0)
  end = map(operator.add, start, source.shape)

  raw_subslices = zip(start, end)
  subslices = slices(raw_subslices)

  destination[subslices] = source



from typing import Callable, Dict, Iterable, Tuple

import operator

from functools import partial

import numpy as np
import pandas as pd

from sklearn.preprocessing import OneHotEncoder

from nonion import Pipeline
from nonion import key
from nonion import mapexplode
from nonion import star
from nonion import value

Encoder = Callable[[Tuple[object, ...]], pd.DataFrame]

COLUMNS_AXIS = 1

COLUMNS_SORT_PARAMETERS = {
  "axis": COLUMNS_AXIS,
  "kind": "stable"
}

ENCODER_PARAMETERS = {
  "sparse": False,
  "handle_unknown": "ignore"
}

def unique(frame: pd.DataFrame) -> Dict[str, Tuple[object, ...]]:
  return (
    Pipeline(frame.iteritems())
    / value(lambda y: y.unique())
    / value(lambda y: y.astype(object))
    / value(tuple)
    >> dict
  )

def create_dummifier(
    xys: Dict[str, Tuple[object, ...]],
    separator: str = "_",
    sort_columns: bool = True
  ) -> Callable[[pd.DataFrame], pd.DataFrame]:

  create: Callable[[str, Tuple[object, ...]], Encoder] = partial(create_encoder, separator=separator)

  name_to_encoder: Dict[str, Encoder] = (
    Pipeline(xys.items())
    / star(lambda x, y: (x, create(x, y)))
    >> dict
  )

  sort = (lambda x: x.sort_index(**COLUMNS_SORT_PARAMETERS)) if sort_columns else (lambda x: x)

  def dummify(frame: pd.DataFrame) -> pd.DataFrame:
    sorted_frame: pd.DataFrame = sort(frame)

    return (
      Pipeline(sorted_frame.iteritems())
      / key(lambda x: name_to_encoder.get(x, pd.DataFrame))
      / star(lambda f, x: f(x))
      >> partial(pd.concat, axis=COLUMNS_AXIS)
    )

  return dummify

def create_encoder(x: str, ys: Tuple[object, ...], separator: str = "_") -> Encoder:
  encoder = OneHotEncoder(**ENCODER_PARAMETERS)

  expanded_ys = np.expand_dims(ys, axis=-1)
  encoder.fit(expanded_ys)

  columns: Iterable[str] = mapexplode(operator.add, x + separator, map(str, ys))
  columns = tuple(columns)

  def encode(zs: Tuple[object, ...]) -> pd.DataFrame:
    expanded_zs = np.expand_dims(zs, axis=-1)
    frame = encoder.transform(expanded_zs)

    return pd.DataFrame(frame, columns=columns)

  return encode



from .tensor import create_batches
from .tensor import cycle_tensor
from .tensor import expand_tail
from .tensor import expand_tail_dimensions
from .tensor import gridrange
from .tensor import iteraxis
from .tensor import move_tensor
from .tensor import move_tensor_inplace
from .tensor import next_batch
from .tensor import slices



import setuptools

with open("README.md") as buffer:
    long_description = buffer.read()

setuptools.setup(
    name="dskit",
    version="0.0.1",
    author="Illia Shkroba",
    author_email="is@pjwstk.edu.pl",
    description="Python Data Science Kit for Humans.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://bitbucket.org/shkroba/dskit",
    packages=setuptools.find_packages(),
    classifiers=[
        "Intended Audience :: Developers",
        "Intended Audience :: Science/Research",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 3.6",
        "Topic :: Software Development :: Libraries :: Application Frameworks",
        "Typing :: Typed"
    ],
    python_requires='>=3.6',
    install_requires=(
      "nonion>=0.0.4",
      "numpy>=1.16",
      "pandas>=1.1.0",
      "scikit-learn>=0.23"
    )
)
